/*
 * Regular expression implementation.
 * Supports only ( | ) * + ?.  No escapes.
 * Compiles to NFA and then simulates NFA
 * using Thompson's algorithm.
 *
 * See also http://swtch.com/~rsc/regexp/ and
 * Thompson, Ken.  Regular Expression Search Algorithm,
 * Communications of the ACM 11(6) (June 1968), pp. 419-422.
 * 
 * Copyright (c) 2007 Russ Cox.
 * Can be distributed under the MIT license, see bottom of file.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "construct.h"

#include <fstream>
#include <list>
#include <string>
#include <iostream>
#include <iterator>
using namespace std;

/**
 * 	typedef struct State State;
 * 	struct State
 * 	{
 * 		int c;
 * 		State *out;
 * 		State *out1;
 * 		int lastlist;
 * 	};
 * 	State matchstate = { Match };	
 **/

typedef struct List List;
struct List
{
	State **s;
	int n;
};
List l1, l2;
static int listid;

void addstate(List*, State*);
void step(List*, int, List*);

/* Compute initial state list */
List* startlist(State *start, List *l)
{
	l->n = 0;
	listid++;
	addstate(l, start);
	return l;
}

/* Check whether state list contains a match. */
int ismatch(List *l)
{
	int i;

	for(i=0; i<l->n; i++)
		if(l->s[i] == &matchstate)
			return 1;
	return 0;
}

/* Add s to l, following unlabeled arrows. */
void addstate(List *l, State *s)
{
	if(s == NULL || s->lastlist == listid)
		return;
	s->lastlist = listid;
	if(s->c == Split){
		/* follow unlabeled arrows */
		addstate(l, s->out);
		addstate(l, s->out1);
		return;
	}
	l->s[l->n++] = s;
}

/*
 * Step the NFA from the states in clist
 * past the character c,
 * to create next NFA state set nlist.
 */
void step(List *clist, int c, List *nlist)
{
	int i;
	State *s;

	listid++;
	nlist->n = 0;
	for(i=0; i<clist->n; i++){
		s = clist->s[i];
		if(s->c == c)
			addstate(nlist, s->out);
	}
}

int matchprefix(State *start, string s, list<string> & prefixes)
{
	List *clist, *nlist, *t;

	clist = startlist(start, &l1);
	nlist = &l2;
	for(int i=0; i<s.length(); ++i)
	{
		int c = (int) s[i];
		step(clist, c, nlist);
		t = clist; clist = nlist; nlist = t;	/* swap clist, nlist */
		if (ismatch(clist))
			prefixes.push_back(s.substr(0,i+1));
	}
	return (! prefixes.empty());	// return true if it is NOT empty
}

int main(int argc, char **argv)
{
	if(argc != 4){
		fprintf(stderr, "usage: nfa regexp.in strings.in accepted.out\n");
		return 1;
	}
	string regfile(argv[1]);
	string strfile(argv[2]);
	string outfile(argv[3]); 

	string regexstr;
	list<string> strs;

	ifstream regexinput(regfile.c_str(), ifstream::in);
	if(! regexinput.is_open())
	{
		cerr << "nfa: can't open " << regfile << endl;
		return -1;
	}
	else
	{
		regexinput >> regexstr; 
		regexinput.close();
	}
	
	ifstream strsinput(strfile.c_str(), ifstream::in);
	if(! strsinput.is_open())
	{
		cerr << "nfa: can't open " << strfile << endl;
		return -1;
	}
	else
	{
		string str;
		while(strsinput >> str)
		{
			strs.push_back(str);
		}
		strsinput.close();
	}

	
	char * post = re2post((char*)regexstr.c_str());
	if(post == NULL)
	{
		cerr << "bad regexp " << regexstr << endl;
		return -1;
	}

	State * start = post2nfa(post);
	if(start == NULL)
	{
		cerr << "error in post2nfai " << post << endl;
		return -1;
	}
	
	l1.s = (State **) malloc(nstate*sizeof(l1.s[0]));
	l2.s = (State **) malloc(nstate*sizeof(l2.s[0]));

	ofstream accepted(outfile.c_str());
	if(! accepted.is_open())
	{
		cerr << "nfa: can't open " << outfile << endl;
		return -1;
	}
	else
	{
		ostream_iterator<string> os( accepted, "\n" );

		for(list<string>::iterator itr=strs.begin(); itr != strs.end(); ++itr)	// not allowed to parallelize
		{
			int strlen = (*itr).length();
			for(int i=0; i< strlen; ++i)	
			{
				list<string> valids;
				string suffix = (*itr).substr(i, strlen-i);
				if(matchprefix(start, suffix, valids))
				{
					copy(valids.begin(), valids.end(), os);
				}
			}
		}
	}
	accepted.close();
	free(l1.s);
	free(l2.s);

	return 0;
}

/*
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated
 * documentation files (the "Software"), to deal in the
 * Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute,
 * sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall
 * be included in all copies or substantial portions of the
 * Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
 * KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE AUTHORS
 * OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
